#pragma once

#include "Exception.h"

namespace MyLib {
class RegistryValue
{
	DWORD type_;
	BYTE *ptr_;
	DWORD size_;

	// Zq
	RegistryValue& operator=(RegistryValue const &right);

public:
	RegistryValue(RegistryValue const &right)
		: type_(right.type_)
		, ptr_(new BYTE[right.size_])
		, size_(right.size_)
	{
		memcpy(this->ptr_, right.ptr_, right.size_);
	}

	RegistryValue(DWORD value_type, void *value_ptr, DWORD value_size)
		: type_(value_type)
		, ptr_(new BYTE[value_size])
		, size_(value_size)
	{
		memcpy(this->ptr_, value_ptr, value_size);
	}

	~RegistryValue()
	{
		delete[] (this->ptr_);
	}

	DWORD getType() const
	{
		return this->type_;
	}

	BYTE const* get() const
	{
		return this->ptr_;
	}

	DWORD getSize() const
	{
		return this->size_;
	}
};

class RegistryKey
{
	HKEY hkey_;

public:
	RegistryKey(HKEY hkey)
		: hkey_(hkey)
	{
	}

	RegistryKey(HKEY parent_key, wchar_t const *key_name, REGSAM sam_desired = KEY_READ | KEY_WRITE)
		: hkey_(NULL)
	{
		long ret = ::RegOpenKeyEx(parent_key, key_name, 0, sam_desired, &(this->hkey_));
		if (ret != ERROR_SUCCESS) {
			BOOST_THROW_EXCEPTION(Win32Exception(ret));
		}
	}

	~RegistryKey()
	{
		if (this->hkey_ != NULL) {
			::RegCloseKey(this->hkey_);
		}
	}

	HKEY get() const
	{
		return this->hkey_;
	}

	HKEY createKey(wchar_t const *key_name, DWORD options = REG_OPTION_NON_VOLATILE,
		REGSAM sam_desired = KEY_READ | KEY_WRITE, SECURITY_ATTRIBUTES *sec_attribs = NULL) const
	{
		return createKey(this->hkey_, key_name, options, sam_desired, sec_attribs);
	}

	static HKEY createKey(HKEY parent_key, wchar_t const *key_name, DWORD options = REG_OPTION_NON_VOLATILE,
		REGSAM sam_desired = KEY_READ | KEY_WRITE, SECURITY_ATTRIBUTES *sec_attribs = NULL)
	{
		HKEY hk;
		long ret = ::RegCreateKeyEx(parent_key, key_name, 0, NULL, options, sam_desired, sec_attribs, &hk, NULL);
		if (ret != ERROR_SUCCESS) {
			BOOST_THROW_EXCEPTION(Win32Exception(ret));
		}
		return hk;
	}

	bool isKeyExists(wchar_t const *key_name) const
	{
		return isKeyExists(this->hkey_, key_name);
	}

	static bool isKeyExists(HKEY parent_key, wchar_t const *key_name)
	{
		HKEY hk;
		long ret = ::RegOpenKeyEx(parent_key, key_name, 0, KEY_READ, &hk);
		if (ret == ERROR_SUCCESS) {
			::RegCloseKey(hk);
			return true;
		} else if (ret != ERROR_FILE_NOT_FOUND) {
			BOOST_THROW_EXCEPTION(Win32Exception(ret));
		}
		return false;
	}

	DWORD getEntryDataType(wchar_t const *entry_name) const
	{
		DWORD data_type;
		long ret = ::RegQueryValueEx(this->hkey_, entry_name, NULL, &data_type, NULL, NULL);
		if (ret != ERROR_SUCCESS) {
			return ULONG_MAX;
		}
		return data_type;
	}

	////////////////////////////////////////////////////////////////////////

	RegistryValue getValue(wchar_t const *entry_name) const
	{
		DWORD data_type, data_size;
		long ret = ::RegQueryValueEx(this->hkey_, entry_name, NULL, &data_type, NULL, &data_size);
		if (ret != ERROR_SUCCESS) {
			BOOST_THROW_EXCEPTION(Win32Exception(ret));
		}
		boost::scoped_array<BYTE> buf(new BYTE[data_size]);
		ret = ::RegQueryValueEx(this->hkey_, entry_name, NULL, NULL,
			buf.get(), &data_size);
		if (ret != ERROR_SUCCESS) {
			BOOST_THROW_EXCEPTION(Win32Exception(ret));
		}
		return RegistryValue(data_type, buf.get(), data_size);
	}

	DWORD getDWORDValue(wchar_t const *entry_name) const
	{
		DWORD data_type, data;
		DWORD data_size = sizeof(DWORD);
		long ret = ::RegQueryValueEx(this->hkey_, entry_name, NULL, &data_type,
			reinterpret_cast<BYTE*>(&data), &data_size);
		if (ret != ERROR_SUCCESS) {
			BOOST_THROW_EXCEPTION(Win32Exception(ret));
		}
		if (data_type != REG_DWORD) {
			BOOST_THROW_EXCEPTION(Win32Exception(ERROR_INVALID_DATA));
		}
		return data;
	}

	std::wstring getStringValue(wchar_t const *entry_name) const
	{
		DWORD data_type, data_size;
		long ret = ::RegQueryValueEx(this->hkey_, entry_name, NULL,
			&data_type, NULL, &data_size);
		if (ret != ERROR_SUCCESS) {
			BOOST_THROW_EXCEPTION(Win32Exception(ret));
		}
		if (data_type != REG_SZ && data_type != REG_EXPAND_SZ) {
			BOOST_THROW_EXCEPTION(Win32Exception(ERROR_INVALID_DATA));
		}
		boost::scoped_array<BYTE> buf(new BYTE[data_size]);
		ret = ::RegQueryValueEx(this->hkey_, entry_name, NULL, NULL,
			buf.get(), &data_size);
		if (ret != ERROR_SUCCESS) {
			BOOST_THROW_EXCEPTION(Win32Exception(ret));
		}
		return reinterpret_cast<wchar_t*>(buf.get());
	}

	////////////////////////////////////////////////////////////////////////

	void setValue(wchar_t const *entry_name, RegistryValue const &value)
	{
		long ret = ::RegSetValueEx(this->hkey_, entry_name, 0, value.getType(), value.get(), value.getSize());
		if (ret != ERROR_SUCCESS) {
			BOOST_THROW_EXCEPTION(Win32Exception(ret));
		}
	}

	void setDWORDValue(wchar_t const *entry_name, DWORD value)
	{
		long ret = ::RegSetValueEx(this->hkey_, entry_name, 0, REG_DWORD,
			reinterpret_cast<BYTE*>(&value), sizeof(DWORD));
		if (ret != ERROR_SUCCESS) {
			BOOST_THROW_EXCEPTION(Win32Exception(ret));
		}
	}

	void setStringValue(wchar_t const *entry_name, wchar_t const *value, bool is_expand_sz = false)
	{
		long ret = ::RegSetValueEx(this->hkey_, entry_name, 0, (is_expand_sz ? REG_EXPAND_SZ : REG_SZ),
			reinterpret_cast<BYTE const *>(value), static_cast<DWORD>((wcslen(value) + 1)*sizeof(wchar_t)));
		if (ret != ERROR_SUCCESS) {
			BOOST_THROW_EXCEPTION(Win32Exception(ret));
		}
	}

	// Query/SetMultiStringValue
	// RecurseDeleteKey
};
}; // namespace MyLib


//class CRegKey
//{
//public:
//	CRegKey() throw();
//	CRegKey( CRegKey& key ) throw();
//	explicit CRegKey(HKEY hKey) throw();
//	~CRegKey() throw();
//
//	CRegKey& operator=( CRegKey& key ) throw();
//
//// Attributes
//public:
//	operator HKEY() const throw();
//	HKEY m_hKey;
//
//private:
//        _ATL_DELETEKEYEXFUNC* pfnDeleteKeyEx;
//        _ATL_DELETEKEYFUNC* pfnDeleteKey;
//
//// Operations
//public:
//	ATL_DEPRECATED("CRegKey::SetValue(DWORD, TCHAR *valueName) has been superseded by CRegKey::SetDWORDValue")
//	LONG SetValue(DWORD dwValue, LPCTSTR lpszValueName);
//	ATL_DEPRECATED("CRegKey::SetValue(TCHAR *value, TCHAR *valueName) has been superseded by CRegKey::SetStringValue and CRegKey::SetMultiStringValue")
//	LONG SetValue(LPCTSTR lpszValue, LPCTSTR lpszValueName = NULL, bool bMulti = false, int nValueLen = -1);
//	LONG SetValue(LPCTSTR pszValueName, DWORD dwType, const void* pValue, ULONG nBytes) throw();
//	LONG SetGUIDValue(LPCTSTR pszValueName, REFGUID guidValue) throw();
//	LONG SetBinaryValue(LPCTSTR pszValueName, const void* pValue, ULONG nBytes) throw();
//	LONG SetDWORDValue(LPCTSTR pszValueName, DWORD dwValue) throw();
//	LONG SetQWORDValue(LPCTSTR pszValueName, ULONGLONG qwValue) throw();
//	LONG SetStringValue(__in_z_opt LPCTSTR pszValueName, __in_z LPCTSTR pszValue, __in DWORD dwType = REG_SZ) throw(...);
//	LONG SetMultiStringValue(LPCTSTR pszValueName, LPCTSTR pszValue) throw(...);
//
//	ATL_DEPRECATED("CRegKey::QueryValue(DWORD, TCHAR *valueName) has been superseded by CRegKey::QueryDWORDValue")
//	LONG QueryValue(__out DWORD& dwValue, __in_z_opt LPCTSTR lpszValueName);
//	ATL_DEPRECATED("CRegKey::QueryValue(TCHAR *value, TCHAR *valueName) has been superseded by CRegKey::QueryStringValue and CRegKey::QueryMultiStringValue")
//	LONG QueryValue(__out_ecount_part_z_opt(*pdwCount, *pdwCount) LPTSTR szValue, __in_z_opt LPCTSTR lpszValueName, __inout DWORD* pdwCount);
//	LONG QueryValue(__in_z_opt LPCTSTR pszValueName, DWORD* pdwType, void* pData, ULONG* pnBytes) throw();
//	LONG QueryGUIDValue(__in_z_opt LPCTSTR pszValueName, GUID& guidValue) throw();
//	LONG QueryBinaryValue(__in_z_opt LPCTSTR pszValueName, void* pValue, ULONG* pnBytes) throw();
//	LONG QueryDWORDValue(__in_z_opt LPCTSTR pszValueName, DWORD& dwValue) throw();
//	LONG QueryQWORDValue(__in_z_opt LPCTSTR pszValueName, ULONGLONG& qwValue) throw();
//	LONG QueryStringValue(__in_z_opt LPCTSTR pszValueName, __out_ecount_part_z_opt(*pnChars, *pnChars) LPTSTR pszValue, __inout ULONG* pnChars) throw();
//
//	LONG QueryMultiStringValue(__in_z_opt LPCTSTR pszValueName, __out_ecount_part_z_opt(*pnChars, *pnChars) __out_z LPTSTR pszValue, __inout ULONG* pnChars) throw();
//
//	// Get the key's security attributes.
//	LONG GetKeySecurity(SECURITY_INFORMATION si, PSECURITY_DESCRIPTOR psd, LPDWORD pnBytes) throw();
//	// Set the key's security attributes.
//	LONG SetKeySecurity(SECURITY_INFORMATION si, PSECURITY_DESCRIPTOR psd) throw();
//
//	LONG SetKeyValue(LPCTSTR lpszKeyName, LPCTSTR lpszValue, LPCTSTR lpszValueName = NULL) throw();
//	static LONG WINAPI SetValue(HKEY hKeyParent, LPCTSTR lpszKeyName,
//		LPCTSTR lpszValue, LPCTSTR lpszValueName = NULL);
//
//	// Create a new registry key (or open an existing one).
//	LONG Create(__in HKEY hKeyParent, __in_z LPCTSTR lpszKeyName,
//		__in_z_opt LPTSTR lpszClass = REG_NONE, __in DWORD dwOptions = REG_OPTION_NON_VOLATILE,
//		__in REGSAM samDesired = KEY_READ | KEY_WRITE,
//		__in_opt LPSECURITY_ATTRIBUTES lpSecAttr = NULL,
//		__in_opt LPDWORD lpdwDisposition = NULL) throw();
//	// Open an existing registry key.
//	LONG Open(HKEY hKeyParent, LPCTSTR lpszKeyName,
//		REGSAM samDesired = KEY_READ | KEY_WRITE) throw();
//	// Close the registry key.
//	LONG Close() throw();
//	// Flush the key's data to disk.
//	LONG Flush() throw();
//
//	// Detach the CRegKey object from its HKEY.  Releases ownership.
//	HKEY Detach() throw();
//	// Attach the CRegKey object to an existing HKEY.  Takes ownership.
//	void Attach(HKEY hKey) throw();
//
//	// Enumerate the subkeys of the key.
//	LONG EnumKey(__in DWORD iIndex, __out_ecount_part_z(*pnNameLength, *pnNameLength) LPTSTR pszName, __inout LPDWORD pnNameLength, __out_opt FILETIME* pftLastWriteTime = NULL) throw();
//	LONG NotifyChangeKeyValue(BOOL bWatchSubtree, DWORD dwNotifyFilter, HANDLE hEvent, BOOL bAsync = TRUE) throw();
//
//	LONG DeleteSubKey(LPCTSTR lpszSubKey) throw();
//	LONG RecurseDeleteKey(LPCTSTR lpszKey) throw();
//	LONG DeleteValue(LPCTSTR lpszValue) throw();
//};
